/*! \file interface.c
 *  \brief Gui Driver.
 *         Dingoo Library.
 *
 *  \version 0.9
 *
 *  Dingoo Interface/Gui driver
 *
 *   Pituka - Dingoo Port
 *  (c) Copyright 2009 David Colmenero (aka D_Skywalk)
 */

#define DINGOO_SDL

#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/poll.h>
#include <string.h>
#include <sys/soundcard.h>
#include <sys/mman.h>

#ifdef DINGOO_SDL
#include <SDL/SDL.h>

SDL_AudioSpec *audio_spec = NULL;
#else

static pthread_t sound_thrd_loop_fill = (pthread_t)0;
static bool sound_running = false;
static bool sfx_thread_paused = true;
#endif

#include "audio.h"
#include "dngutils.h"


//punteros al sistema de sonido del pituka
extern unsigned char *pbSndBuffer;
extern unsigned char *pbSndBufferEnd;
extern unsigned char *pbSndStream;



int sound_dev = -1;
int mixer_dev = -1;

int vol_left = 0, vol_right = 0;

audio_buf_info sound_buf_info;

int sound_fmt, sound_chans;
long sound_rate;

#define SFX_THREAD_FRAG_SIZE 1024

unsigned char *sound_buffer;

int sound_set_volume() {
	if (vol_left < 0) vol_left = 0;
	if (vol_left > 100) vol_left = 100;
	if (vol_right < 0) vol_right = 0;
	if (vol_right > 100) vol_right = 100;

	int vol = ((vol_left << 8) | vol_right);
	return (ioctl(mixer_dev, MIXER_WRITE(SOUND_MIXER_VOLUME), &vol) == -1
		|| ioctl(mixer_dev, MIXER_WRITE(SOUND_MIXER_PCM), &vol) == -1) ? -1 : 0;
}


int sound_reset() {
	return ioctl(sound_dev, SNDCTL_DSP_RESET, 0);
}

int sound_set_format(int fmt) {
	if (fmt != AFMT_S16_LE) return -1; /* sorry, but this library depends on it */

	sound_reset();

	if (ioctl(sound_dev, SNDCTL_DSP_SETFMT, &fmt) == -1) return -1;
	if (ioctl(sound_dev, SNDCTL_DSP_GETOSPACE, &sound_buf_info) == -1) return -1;

	sound_fmt = fmt;

	return 0;
}

int sound_set_channels(int chans) {
	if (sound_chans == chans) return 0;

	sound_reset();

	if (ioctl(sound_dev, SNDCTL_DSP_CHANNELS, &chans) == -1) return -1;
	if (ioctl(sound_dev, SNDCTL_DSP_GETOSPACE, &sound_buf_info) == -1) return -1;

	sound_chans = chans;

	return 0;
}

int sound_set_rate(long rate) {
	if (sound_rate == rate) return 0;

	sound_reset();

	if (ioctl(sound_dev, SNDCTL_DSP_SPEED, &rate) == -1) return -1;
	if (ioctl(sound_dev, SNDCTL_DSP_GETOSPACE, &sound_buf_info) == -1) return -1;

	sound_rate = rate;

	return 0;
}

int sound_set_buffer_size(int max_frags, int size) {
	sound_reset();

	int frag = (max_frags << 16) | size;

//        printf("AUDIO - INIT: %i\n", frag);

	if (ioctl(sound_dev, SNDCTL_DSP_SETFRAGMENT, &frag) == -1) return -1;
	if (ioctl(sound_dev, SNDCTL_DSP_GETOSPACE, &sound_buf_info) == -1) return -1;

        //sound_buffer = mmap(NULL, SFX_THREAD_FRAG_SIZE << 4, PROT_READ | PROT_WRITE, MAP_SHARED, sound_dev, 0);

	return 0;
}

int sound_get_buffer_remaining() {
	audio_buf_info info;
	if (ioctl(sound_dev, SNDCTL_DSP_GETOSPACE, &info) == -1) return 0;

	return sound_buf_info.bytes - info.bytes;
}

int sound_set_nonblock() {
	return ioctl(sound_dev, SNDCTL_DSP_NONBLOCK, 0);
}

#ifndef DINGOO_SDL

static void *sound_loop_fill(void *user_data) {
	//char buf[1024*2];

	while (sound_running) {
		//memset(buf, 0, sizeof(buf));

		//sound_callback(sound_callback_user_data, buf, sizeof(buf));

		struct pollfd fds[1];

		fds[0].fd = sound_dev;
		fds[0].events = POLLOUT;

		while (!poll(fds, sizeof(fds)/sizeof(fds[0]), 0)) usleep(1000);

		if (sfx_thread_paused) continue;

		if((pbSndStream + SFX_THREAD_FRAG_SIZE) >= pbSndBufferEnd){
                        //ultra_memcpy(sound_buffer, pbSndStream, (pbSndBufferEnd-pbSndStream));
			write(sound_dev, pbSndStream, (pbSndBufferEnd-pbSndStream));
                    	pbSndStream = pbSndBuffer;           // vuelve al comienzo
		}else{
                        //ultra_memcpy(sound_buffer, pbSndStream, SFX_THREAD_FRAG_SIZE);
			write(sound_dev, pbSndStream, SFX_THREAD_FRAG_SIZE);
                	pbSndStream += (SFX_THREAD_FRAG_SIZE); // se prepara para el próximo copiado
		}


	}

	return NULL;
}


int SoundSetup (void)
{

    if ((sound_dev = open("/dev/dsp", O_WRONLY)) == -1) goto error;

    if (sound_set_buffer_size(3, 15)) goto error; /* these values are based on several tests; not accurate */

    if (sound_set_format(AFMT_S16_LE) || sound_set_channels(2) || sound_set_rate(44100)) goto error; /* FIXME */

    //sound_callback = callback;
    //sound_callback_user_data = user_data;

    sound_running = true;
    sfx_thread_paused = false;

    if (!sound_thrd_loop_fill && pthread_create(&sound_thrd_loop_fill, NULL, sound_loop_fill, NULL)) goto error;

    if ((mixer_dev = open("/dev/mixer", O_RDONLY)) == -1) goto error;

    return SFX_THREAD_FRAG_SIZE;// << 1;

error:
    SoundClose();
    return 0;

}


void SoundInit (void){ }


void StopSound ( int val ) {  /*sfx_thread_paused = val;*/ }


void SoundClose(void)
{
	sound_running = false;

	if (sound_thrd_loop_fill && pthread_join(sound_thrd_loop_fill, NULL)) return;
	sound_thrd_loop_fill = (pthread_t)0;

	if (sound_dev != -1 && (close(sound_dev) || ((sound_dev = -1) & 0))) return;

	if (mixer_dev != -1 && (close(mixer_dev) || ((mixer_dev = -1) & 0))) return;

}

#endif

#ifdef DINGOO_SDL

int audio_align_samples (int given)
{
   int actual = 1;

   while (actual < given) {
      actual <<= 2;
   }

   return actual; // return the closest match as 2^n
}

void StopSound ( int val ) {      SDL_PauseAudio(val);  }

void SoundUpdate (void *userdata, unsigned char *stream, int len)
{

   ultra_memcpy(stream, pbSndStream, len);

   pbSndStream += len;
   if (pbSndStream >= pbSndBufferEnd) {
      pbSndStream = pbSndBuffer;
   }

}

int SoundSetup (void)
{
   SDL_AudioSpec *desired, *obtained;

   desired = (SDL_AudioSpec *)malloc(sizeof(SDL_AudioSpec));
   obtained = (SDL_AudioSpec *)malloc(sizeof(SDL_AudioSpec));

   desired->freq = 44100;
   desired->format = AUDIO_S16;
   desired->channels = 2;
   desired->samples = audio_align_samples(desired->freq / 50); // desired is 20ms at the given frequency //367
   desired->callback = SoundUpdate;
   desired->userdata = NULL;

   if (SDL_OpenAudio(desired, obtained) < 0) {
      fprintf(stderr, "Could not open audio: %s\n", SDL_GetError());
      return 0;
   }

   free(desired);
   audio_spec = obtained;

   printf("AUDIO - INIT: %i\n",audio_spec->size);

   return audio_spec->size; // size is samples * channels * bytes per sample (1 or 2)

}

void SoundClose (void)
{

   SDL_CloseAudio();
   if (audio_spec) {
      free(audio_spec);
   }

}

#endif


